package org.kantega.jexmec.services;

/*
 * Copyright 2011 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.ServiceKey;
import org.kantega.jexmec.ServiceLocator;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.lang.reflect.Type;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * Utility class for validating service interfaces and returning service keys for a service interface.
 */
public class Services {

    /**
     * Get the service keys from a given service interface. This method will first validate the services interface
     * by calling {@link #validateServiceInterface(Class)}
     *
     * @param servicesInterface the service interface
     * @return a set of service keys
     */
    public static Set<ServiceKey> getServiceKeysFromInterface(Class servicesInterface) {

        validateServiceInterface(servicesInterface);

        Set<ServiceKey> serviceClasses = new HashSet<ServiceKey>();

        for (Method method : servicesInterface.getDeclaredMethods()) {
            serviceClasses.addAll(Services.getServiceKeysForMethod(method));
        }

        return serviceClasses;
    }

    /**
     * Validate that the given class follows all conventions for being used as a Jexmec service interface.
     * <p/>
     * The class must adhere to the following rules:
     * <p/>
     * <ul>
     * <li>It must be an <code>interface</code> (not a concrete class);
     * <li>Methods cannot return any primitive types</li>
     * <li>Methods cannot return <code>void</code></li>
     * <li>Methods must either have zero parameters or a single <code>enum</code> parameter for named services.
     * <li>No two methods can have the same return type.
     * </ul>
     *
     * @param servicesInterface the class to validate
     */
    public static void validateServiceInterface(Class servicesInterface) {
        if (!servicesInterface.isInterface()) {
            throw new IllegalArgumentException("The class " + servicesInterface + " is not an interface");
        }

        Set<Type> returnTypes = new HashSet<Type>();

        for (Method method : servicesInterface.getDeclaredMethods()) {
            if (method.getParameterTypes().length > 1) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is only allowed to take a single naming enumeration as parameter. " +
                        "Please remove parameters " + Arrays.asList(method.getParameterTypes()));
            }
            if (method.getParameterTypes().length == 1) {
                if (!method.getParameterTypes()[0].isEnum()) {
                    throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " a naming enumeration as a parameter. " +
                            "Please remove or fix non enum parameter " + method.getParameterTypes()[0]);
                }
            }
            if (method.getReturnType() == void.class) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "." + method.getName() + " is not allowed to return void");
            }
            if (method.getReturnType().isPrimitive()) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "."
                        + method.getName() + " is not allowed to return primitive types. The method returns "
                        + method.getReturnType());
            }
            if (returnTypes.contains(method.getReturnType())) {
                throw new IllegalArgumentException("The service method " + servicesInterface.getName() + "."
                        + method.getName() + " returns " + method.getReturnType() + " which has already been used in this service interface");
            } else {
                returnTypes.add(method.getReturnType());
            }
        }
    }

    /**
     * Returns the set of service keys for a single service interface method.
     *
     * @param method the method to return service keys for
     * @return the service keys
     */
    public static Set<ServiceKey> getServiceKeysForMethod(Method method) {
        if (method.getParameterTypes().length == 1) {
            Set<ServiceKey> serviceKeys = new HashSet<ServiceKey>();
            Object[] constants = method.getParameterTypes()[0].getEnumConstants();
            for (Object object : constants) {
                if (object instanceof Enum) {
                    serviceKeys.add(ServiceKey.by(method.getReturnType(), Enum.class.cast(object).name()));
                }
            }
            return serviceKeys;
        } else {
            return Collections.<ServiceKey>singleton(ServiceKey.by(method.getReturnType()));
        }
    }

    /**
     * Creates an instance of a services interface. The instance is implemented as
     * a dynamic proxy that will look up services the service locators added to this plugin manager
     *
     * @param servicesClass the service class to implement
     * @param <S>           the service class type
     * @return an instance of the service interface
     */
    public static <S> S createServicesInstance(final Class<S> servicesClass, final ServiceLocator serviceLocator) {
        return servicesClass.cast(Proxy.newProxyInstance(servicesClass.getClassLoader(),
                new Class[]{servicesClass},
                new InvocationHandler() {
                    public Object invoke(Object o, Method method, Object[] objects) throws Throwable {
                        ServiceKey<?> key = objects == null ?
                                ServiceKey.by(method.getReturnType()) : ServiceKey.by(method.getReturnType(), ((Enum) objects[0]).name());

                        return serviceLocator.get(key);
                    }
                }));
    }
}
